@@ -268,17 +268,20 @@ static void crawler_metadump_eval(crawler_module_t *cm, item *it, uint32_t hv, i
         refcount_decr(it);
         return;
     }
+    client_flags_t flags;
+    FLAGS_CONV(it, flags);
     // TODO: uriencode directly into the buffer.
     uriencode(ITEM_key(it), keybuf, it->nkey, KEY_MAX_URI_ENCODED_LENGTH);
     int total = snprintf(cm->c.buf + cm->c.bufused, 4096,
-            "key=%s exp=%ld la=%llu cas=%llu fetch=%s cls=%u size=%lu\n",
+            "key=%s exp=%ld la=%llu cas=%llu fetch=%s cls=%u size=%lu flags=%llu\n",
             keybuf,
             (it->exptime == 0) ? -1 : (long)(it->exptime + process_started),
             (unsigned long long)(it->time + process_started),
             (unsigned long long)ITEM_get_cas(it),
             (it->it_flags & ITEM_FETCHED) ? "yes" : "no",
             ITEM_clsid(it),
-            (unsigned long) ITEM_ntotal(it));
+            (unsigned long) ITEM_ntotal(it),
+            (unsigned long long)flags);
     refcount_decr(it);
     // TODO: some way of tracking the errors. these should be impossible given
     // the space requirements.
