@@ -22,6 +22,7 @@ enum log_entry_type {
     LOGGER_SLAB_MOVE,
     LOGGER_CONNECTION_NEW,
     LOGGER_CONNECTION_CLOSE,
+    LOGGER_DELETIONS,
 #ifdef EXTSTORE
     LOGGER_EXTSTORE_WRITE,
     LOGGER_COMPACT_START,
@@ -108,6 +109,14 @@ struct logentry_item_store {
     char key[];
 };
 
+struct logentry_deletion {
+    int nbytes;
+    int cmd;
+    uint8_t nkey;
+    uint8_t clsid;
+    char key[];
+};
+
 struct logentry_conn_event {
     int transport;
     int reason;
@@ -126,6 +135,17 @@ struct logentry_proxy_req {
     long elapsed;
     char data[];
 };
+
+struct logentry_proxy_errbe {
+    size_t errlen;
+    size_t be_namelen;
+    size_t be_portlen;
+    size_t be_labellen;
+    size_t be_rbuflen;
+    int be_depth;
+    int retry;
+    char data[];
+};
 #endif
 /* end intermediary structures */
 
@@ -155,6 +175,7 @@ struct _logentry {
 #define LOG_PROXYREQS  (1<<10) /* command logs from proxy */
 #define LOG_PROXYEVENTS (1<<11) /* error log stream from proxy */
 #define LOG_PROXYUSER (1<<12) /* user generated logs from proxy */
+#define LOG_DELETIONS (1<<13) /* see whats deleted */
 
 typedef struct _logger {
     struct _logger *prev;
